"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame
from numpy import random

from gym import error, spaces, utils
from gym.utils import seeding

class Bandits(gym.Env):
    def __init__(self, 
            arms = 10,
            max_steps = 5000,
            ):

        self.max_steps = max_steps

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Discrete(arms)

        # observation is the x, y coordinate of the grid
        self.observation_space = None 

        # expected ctr of each arms 
        self.exp_gains = None
        self.K = arms
        self.need_reset = True

        assert self.K > 1 and self.max_steps > 1

    def sample_task(self,
            distribution_settings="Classical",
            mean=0.50,
            dev=0.05,
            ):
        """
        Classical: 1 arm is mean + sqrt(K-1) dev, others are mean - dev / sqrt(K-1)
        Uniform:   each arm with expected gain being Uniform(mean - 1.732 * dev, mean + 1.732 * dev)
        Gaussian:   Gaussian(mean, dev)
        """
        if(distribution_settings == "Classical"):
            fac = numpy.sqrt(self.K - 1)
            exp_gains = numpy.full((self.K,), mean - dev / fac)
            sel_idx = random.randint(0, self.K - 1)
            exp_gains[sel_idx] = mean + fac * dev
            exp_gains = numpy.clip(exp_gains, 0.0, 1.0)
        elif(distribution_settings == "Uniform"):
            exp_gains = numpy.clip((random.random(shape=(self.K,)) - 0.50) * 3.464 + mean, 0.0, 1.0).tolist()
        elif(distribution_settings == "Gaussian"):
            exp_gains = numpy.clip(random.normal(loc=mean, scale=dev), 0.0, 1.0)
        else:
            raise Exception("No such distribution_settings: %s", distribution_settings)
        return exp_gains

    def set_task(self, task_config):
        self.exp_gains = task_config
        assert numpy.shape(self.exp_gains) == (self.K,)
        self.need_reset = True

    def reset(self):
        if(self.exp_gains is None):
            raise Exception("Must call \"set_task\" before reset")
        self.steps = 0
        self.need_reset = False
        return None

    def step(self, action):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        exp_gain = self.exp_gains[action]
        reward = 1 if random.random() < exp_gain else 0
        info = {"steps": self.steps, "expected_gain": exp_gain}
        self.steps += 1
        done = self.steps >= self.max_steps
        if(done):
            self.need_reset = True

        return None, reward, done, info

    def expected_upperbound(self):
        return self.max_steps * numpy.max(self.exp_gains)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gym.envs.registration import register
from l3c.mazeworld.envs import MazeWorldContinuous3D
from l3c.mazeworld.envs import MazeWorldDiscrete3D
from l3c.mazeworld.envs import MazeWorldDiscrete2D
from l3c.mazeworld.envs import MazeTaskSampler

register(
    id='mazeworld-continuous-3D-v1',
    entry_point='l3c.mazeworld:MazeWorldContinuous3D',
    kwargs={
        "enable_render": True,
        "render_scale": 480,
        "resolution": (256, 256),
        "max_steps": 5000,
        "visibility_3D": 12.0,
        "task_type": "NAVIGATION"
    }
)

register(
    id='mazeworld-discrete-3D-v1',
    entry_point='l3c.mazeworld:MazeWorldDiscrete3D',
    kwargs={
        "enable_render": True,
        "render_scale": 480,
        "resolution": (256, 256),
        "max_steps": 2000,
        "visibility_3D": 12.0,
        "task_type": "NAVIGATION"
    }
)

register(
    id='mazeworld-discrete-2D-v1',
    entry_point='l3c.mazeworld:MazeWorldDiscrete2D',
    kwargs={
        "enable_render": True,
        "max_steps": 2000,
        "visibility_2D": 1,
        "task_type": "NAVIGATION"
    }
)



#   Copyright (c) 2021 DeepEvolution Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import _io
import numpy
import gym
from numpy import random

class RandomNGram(object):
    def __init__(self, batch=1, n_emb=16, n_hidden=64, n_vocab=256, n_gram=3, hardness=5, seed=None):
        self.n_emb = n_emb
        self.n_hidden = n_hidden
        self.n_vocab = n_vocab
        self.batch = batch
        self.emb = numpy.random.normal(0, 1.0, size=(self.batch, self.n_vocab, self.n_emb))
        self.n_gram = n_gram
        weights = []
        if(seed is not None):
            numpy.random.seed(seed)
        self.weight_i = numpy.random.normal(0, 1.0, size=(self.batch, self.n_gram, self.n_emb, self.n_hidden))
        self.bias_i = numpy.random.normal(0, 1.0, size=(self.batch, self.n_gram, 1, self.n_hidden))
        self.weight_o = numpy.random.normal(0, 1.0, size=(self.batch, self.n_hidden, self.n_vocab))
        self.bias_o = numpy.random.normal(0, 1.0, size=(self.batch, 1, self.n_vocab))
        self.w_arr = numpy.expand_dims(numpy.arange(self.n_gram), axis=[0, 2, 3])
        self.hardness = hardness

        self.s_tok = 0

    def softmax(self, x):
        e_x = numpy.exp(x - numpy.max(x, axis=-1, keepdims=True))
        return e_x / e_x.sum(axis=-1, keepdims=True)

    def forward(self, l):
        ind = 0

        def mean_var_norm(i):
            m_i = numpy.mean(i)
            m_ii = numpy.mean(i * i)
            std = numpy.sqrt(m_ii - m_i * m_i)
            return (1.0 / std) * (i - m_i)

        cur_tok = numpy.full((self.batch,), self.s_tok)
        idxes = numpy.arange(self.batch)
        pad_emb = numpy.expand_dims(numpy.array(self.emb[idxes, cur_tok]), axis=1)
 
        h = numpy.zeros((self.batch, self.n_hidden))

        # mark whether there is end token
        seqs = []
        seqs.append(cur_tok)
        ppl = 0
        tok_cnt = 0
        tok_embs = [pad_emb for _ in range(self.n_gram)]
        while ind < l:
            ind += 1
            tok_emb = numpy.expand_dims(numpy.array(self.emb[idxes, cur_tok]), axis=1)
            tok_embs.append(tok_emb)
            del tok_embs[0]
            tok_emb = numpy.expand_dims(numpy.concatenate(tok_embs[-self.n_gram:], axis=1), axis=2)

            h = numpy.tanh(numpy.matmul(tok_emb, self.weight_i) + self.bias_i)
            h = numpy.mean(self.w_arr * h, axis=1)
            o = numpy.matmul(h, self.weight_o) + self.bias_o
            o = numpy.squeeze(o, axis=1)
            o = self.hardness * mean_var_norm(o)
            exp_o = numpy.exp(o)
            prob = exp_o / numpy.sum(exp_o, axis=-1, keepdims=True)
            cur_tok = (prob.cumsum(1) > numpy.random.rand(prob.shape[0])[:,None]).argmax(1)
            cur_prob = prob[idxes, cur_tok]
            ppl -= numpy.sum(numpy.log(cur_prob))
            tok_cnt += cur_prob.shape[0]

            seqs.append(cur_tok)
        print("GT Perplexity: %f" % (ppl / tok_cnt))

        return numpy.transpose(numpy.asarray(seqs, dtype="int32"))

class MetaLangV2(gym.Env):
    """
    Pseudo Langauge Generated from RNN models
    V: vocabulary size
    d: embedding size (input size)
    n: n-gram
    N: hidden size
    e: inverse of softmax - temporature
    L: maximum length
    """
    def __init__(self, 
            V=64, 
            n=3,
            d=16,
            N=64,
            e=0.20,
            L=4096):
        self.L = int(L)
        self.V = int(V)
        self.n = n
        self.d = d
        self.N = N
        self.hardness = 1.0/e
        assert n > 1 and V > 1 and N > 1 and L > 1 

    def data_generator(self, seed=None):
        nn = RandomNGram(n_emb = self.d, n_gram=self.n, n_hidden = self.N, n_vocab = self.V, hardness=self.hardness, seed=seed)
        tokens = nn.forward(self.L)[0]
        return tokens

    def batch_generator(self, batch_size, seed=None):
        nn = RandomNGram(batch = batch_size, n_emb = self.d, n_gram=self.n, n_hidden = self.N, n_vocab = self.V, hardness=self.hardness, seed=seed)
        tokens = nn.forward(self.L)
        return tokens

    def generate_text(self, size, output_stream):
        tokens = self.batch_generator(size)
        if(isinstance(output_stream, _io.TextIOWrapper)):
            need_close = False
        elif(isinstance(output_stream, str)):
            output_stream = open(output_stream, "w")
            need_close = True
        for i in range(tokens.shape[0]):
            output_stream.write("\t".join(map(str, tokens[i].tolist())))
            output_stream.write("\n")
        if(need_close):
            output_stream.close()

    def generate_npy(self, size, file_name):
        tokens = self.batch_generator(size)
        numpy.save(file_name, tokens)

    @property
    def VocabSize(self):
        return self.V

    @property
    def SepID(self):
        raise Exception("Not Defined")

    @property
    def MaskID(self):
        return -1

    @property
    def PaddingID(self):
        return -1



#   Copyright (c) 2021 DeepEvolution Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is used to generate data for meta language models

import sys
import argparse
from l3c.metalang import MetaLangV1
from l3c.metalang import MetaLangV2

if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Generating Pseudo-Training Data')
    parser.add_argument('--version', type=str, choices=['v1', 'v2'], default='v2')
    parser.add_argument('--vocab_size', type=int, default=64)
    parser.add_argument('--embedding_size', type=int, default=16)
    parser.add_argument('--hidden_size', type=int, default=16)
    parser.add_argument('--elements_length', type=int, default=64)
    parser.add_argument('--elements_number', type=int, default=10)
    parser.add_argument('--error_rate', type=float, default=0.20)
    parser.add_argument('--n_gram', type=float, default=3)
    parser.add_argument('--sequence_length', type=int, default=4096)
    parser.add_argument('--samples', type=int, default=100)
    parser.add_argument('--output_type', type=str, choices=['txt', 'npy'], default='txt')
    parser.add_argument('--output', type=str, default=None)

    args = parser.parse_args()

    if(args.version == 'v1'):
        dataset = MetaLangV1(
                V=args.vocab_size,
                n=args.elements_number,
                l=args.elements_length,
                e=args.error_rate,
                L=args.sequence_length)
    elif(args.version == 'v2'):
        dataset = MetaLangV2(
                V=args.vocab_size,
                n=args.n_gram,
                d=args.embedding_size,
                N=args.hidden_size,
                e=args.error_rate,
                L=args.sequence_length)

    if(args.output_type == 'npy'):
        if(args.output is None):
            raise Exception("Must specify --output when output_type is npy")
        dataset.generate_npy(args.samples, args.output)
    elif(args.output_type == 'txt'):
        if(args.output is None):
            dataset.generate_text(args.samples, sys.stdout)
        else:
            dataset.generate_text(args.samples, args.output)



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
from pygame import font
from numpy import random as npyrnd
from numpy.linalg import norm
from l3c.mazeworld.envs.maze_base import MazeBase
from .ray_caster_utils import landmarks_rgb,landmarks_color

class MazeCoreDiscrete2D(MazeBase):
    def __init__(self, visibility_2D=1, task_type="SURVIVAL", max_steps=5000):
        super(MazeCoreDiscrete2D, self).__init__(
                visibility_2D=visibility_2D,
                task_type=task_type,
                max_steps=max_steps
                )

    def do_action(self, action):
        assert numpy.shape(action) == (2,)
        assert abs(action[0]) < 2 and abs(action[1]) < 2
        tmp_grid_i = self._agent_grid[0] + action[0]
        tmp_grid_j = self._agent_grid[1] + action[1]

        if(self._cell_walls[tmp_grid_i, tmp_grid_j] < 1):
            self._agent_grid[0] = tmp_grid_i
            self._agent_grid[1] = tmp_grid_j
        self._agent_loc = self.get_cell_center(self._agent_grid)

        reward, done = self.evaluation_rule()
        self.update_observation()
        return reward, done

    def render_observation(self):
        #Paint Observation
        empty_range = 40
        obs_surf = pygame.surfarray.make_surface(self._observation)
        obs_surf = pygame.transform.scale(obs_surf, (self._view_size - 2 * empty_range, self._view_size - 2 * empty_range))
        self._screen.blit(self._obs_logo,(5, 5))
        self._screen.blit(obs_surf, (empty_range, empty_range))

        # Paint the blue edge for observation
        pygame.draw.rect(self._screen, pygame.Color("blue"), 
                (empty_range, empty_range,
                self._view_size - 2 * empty_range, self._view_size - 2 * empty_range), width=1)
        # Paint agent in god view map
        pygame.draw.circle(self._screen, pygame.Color("gray"), 
                ((self._agent_grid[0] + 0.5) * self._render_cell_size + self._view_size, (self._agent_grid[1] + 0.5) * self._render_cell_size),
                int(0.40 * self._render_cell_size), width=0)

    def movement_control(self, keys):
        #Keyboard control cases
        if keys[pygame.K_LEFT]:
            return (-1, 0)
        if keys[pygame.K_RIGHT]:
            return (1, 0)
        if keys[pygame.K_UP]:
            return (0, 1)
        if keys[pygame.K_DOWN]:
            return (0, -1)
        if keys[pygame.K_SPACE]:
            return (0, 0)
        return None

    def update_observation(self):
        self._observation = self.get_loc_map(self.visibility_2D)

        if(self.task_type == "SURVIVAL"):
            # For survival task, the color of the center represents its life value
            f = max(0, int(255 - 128 * self._life))
            self._observation[self.visibility_2D, self.visibility_2D] = numpy.asarray([255, f, f], dtype="int32")
        elif(self.task_type == "NAVIGATION"):
            # For navigation task, the color of the center represents the navigation target
            self._observation[self.visibility_2D, self.visibility_2D] = landmarks_rgb[self._command]

        self._cell_exposed = numpy.zeros_like(self._cell_walls).astype(numpy.bool)
        self._cell_exposed[(self._agent_grid[0] - self.visibility_2D) : (self._agent_grid[0] + self.visibility_2D + 1), \
                (self._agent_grid[1] - self.visibility_2D) : (self._agent_grid[1] + self.visibility_2D + 1)] = True



import numpy

def arr(size):
    x = numpy.random.normal(0, 1.0, size=(size,))
    factor = 1 - 1 / (size ** 0.5) + numpy.log(size)
    x = numpy.exp(x * factor)
    return x / numpy.sum(x)

def ent(x):
    return - numpy.sum(x * numpy.log(x))

def stat_ent(size, stat=1024):
    ents = 0
    print("factor", (numpy.log(size) + 1 / numpy.sqrt(size)))
    for _ in range(stat):
        ents += ent(arr(size))
    return ents / stat

print(stat_ent(8))
print(stat_ent(16))
print(stat_ent(32))
print(stat_ent(64))
print(stat_ent(128))
print(stat_ent(256))
print(stat_ent(4096))
print(stat_ent(10000))
print(stat_ent(40000))



"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame

from gym import error, spaces, utils
from gym.utils import seeding
from l3c.mazeworld.envs.maze_discrete_2d import MazeCoreDiscrete2D
from l3c.mazeworld.envs.maze_continuous_3d import MazeCoreContinuous3D
from l3c.mazeworld.envs.maze_discrete_3d import MazeCoreDiscrete3D

class MazeWorldEnvBase(gym.Env):
    """
    All Maze World Environments Use This Base Class
    """
    def __init__(self, 
            maze_type,
            enable_render=True,
            render_scale=480,
            max_steps=5000,
            task_type="NAVIGATION",
            ):
        self.maze_type = maze_type
        self.enable_render = enable_render
        self.render_viewsize = render_scale
        self.task_type = task_type

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config)
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def action_control(self, action):
        raise NotImplementedError("Must implement the action control logic")

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")

        internal_action = self.action_control(action)
            
        # In keyboard control, process only continues when key is pressed
        info = {"steps": self.maze_core.steps}
        if(internal_action is None):
            return self.maze_core.get_observation(), 0, False, info 
        reward, done = self.maze_core.do_action(internal_action)

        if(done):
            self.need_reset=True

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        if(self.enable_render):
            self.key_done, self.keyboard_press = self.maze_core.render_update()

    def get_loc_map(self, map_range=2):
        return self.maze_core.get_loc_map(map_rang=map_range)

    def get_target_gt_direction(self):
        if(self.task_type != "NAVIGATION"):
            raise Exception("Only \"NAVIGATION\" task type is supported")
        target_id = self.maze_core._commands_sequence[self.maze_core._commands_sequence_idx]
        target_grid = self.maze_core._landmarks_coordinates[target_id]
        deta_grid = numpy.zeros(shape=(2,), dtype=numpy.float32)
        deta_grid[0] = target_grid[0] - self.maze_core._agent_grid[0]
        deta_grid[1] = target_grid[1] - self.maze_core._agent_grid[1]
        angle = numpy.arctan2(deta_grid[1], deta_grid[0]) - self.maze_core._agent_ori
        if(angle < -numpy.pi):
            angle += 2 * numpy.pi
        elif(angle > numpy.pi):
            angle -= 2 * numpy.pi
        return angle

    def save_trajectory(self, file_name, view_size=480):
        if(not self.enable_render):
            self.maze_core.render_init(view_size)
        self.maze_core.render_trajectory(file_name)


class MazeWorldDiscrete3D(MazeWorldEnvBase):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            max_steps=5000,
            task_type="NAVIGATION",
            resolution=(320, 320),
            visibility_3D=12.0,
            ):
        super(MazeWorldDiscrete3D, self).__init__(
            "Discrete3D",
            enable_render=enable_render,
            render_scale=render_scale,
            max_steps=max_steps,
            task_type=task_type
        )
        self.maze_core = MazeCoreDiscrete3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                visibility_3D=visibility_3D,
                task_type = task_type,
                )
        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Discrete(5)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 255, dtype=numpy.float32),
                dtype=numpy.float32)
        self.discrete_actions=[(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1)]

    def action_control(self, action):
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            return self.maze_core.movement_control(self.keyboard_press)
        else:
            return self.discrete_actions[action]

class MazeWorldContinuous3D(MazeWorldEnvBase):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            max_steps = 5000,
            task_type = "NAVIGATION",
            resolution=(320, 320),
            visibility_3D=12.0,
            ):
        super(MazeWorldContinuous3D, self).__init__(
            "Continuous3D",
            enable_render=enable_render,
            render_scale=render_scale,
            max_steps=max_steps,
            task_type=task_type
        )
        self.maze_core = MazeCoreContinuous3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                visibility_3D=visibility_3D,
                task_type = task_type
                )

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Box(low=numpy.array([-1.0, -1.0]), 
                high=numpy.array([1.0, 1.0]), dtype=numpy.float32)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 256, dtype=numpy.float32),
                dtype=numpy.float32)

    def action_control(self, action):
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(20) # 50 FPS
            tr, ws = self.maze_core.movement_control(self.keyboard_press)
        else:
            tr = action[0]
            ws = action[1]
        if(tr is None or ws is None):
            return None
        return [tr, ws]

class MazeWorldDiscrete2D(MazeWorldEnvBase):
    def __init__(self,
            enable_render=True,
            render_scale=480,
            max_steps = 5000,
            task_type = "NAVIGATION",
            visibility_2D = 1):
        super(MazeWorldDiscrete2D, self).__init__(
            "Discrete2D",
            enable_render=enable_render,
            render_scale=render_scale,
            max_steps=max_steps,
            task_type=task_type
            )
        self.maze_core = MazeCoreDiscrete2D(visibility_2D=visibility_2D, max_steps=max_steps, task_type=task_type)

        # Go EAST/WEST/SOUTH/NORTH
        self.action_space = spaces.Discrete(4)
        # observation is the x, y coordinate of the grid
        n_w = 2 * visibility_2D + 1
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(n_w, n_w, 3), dtype=numpy.float32), 
                high=numpy.full((n_w, n_w, 3), 255, dtype=numpy.float32),
                dtype=numpy.float32)

        self.discrete_actions=[(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1)]


    def action_control(self, action):
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            return self.maze_core.movement_control(self.keyboard_press)
        else:
            return self.discrete_actions[action]



import numpy
from numba import njit

PI_4 = 0.7853981
PI_2 = 1.5707963
PI = 3.1415926
t_PI = 6.2831852
PI2d = 57.29578

OFFSET_10 = numpy.asarray([0.5, 0.5], dtype="float32")
OFFSET_01 = numpy.asarray([-0.5, 0.5], dtype="float32")
OFFSET_m0 = numpy.asarray([-0.5, -0.5], dtype="float32")
OFFSET_0m = numpy.asarray([0.5, -0.5], dtype="float32")


@njit(cache=True)
def nearest_point(pos, line_1, line_2):
    unit_ori = line_2 - line_1
    edge_norm = numpy.sqrt(numpy.sum(unit_ori * unit_ori))
    unit_ori /= max(1.0e-6, edge_norm)

    dist_1 = numpy.sum((pos - line_1) * unit_ori)
    if(dist_1 > edge_norm):
        return numpy.sqrt(numpy.sum((pos - line_2) ** 2)), numpy.copy(line_2)
    elif(dist_1 < 0):
        return numpy.sqrt(numpy.sum((pos - line_1) ** 2)), numpy.copy(line_1)
    else:
        line_p = line_1 + dist_1 * unit_ori
        return numpy.sqrt(numpy.sum((pos - line_p) ** 2)), numpy.copy(line_p)

@njit(cache=True)
def collision_force(dist_vec, cell_size, col_dist):
    dist = float(numpy.sqrt(numpy.sum(dist_vec * dist_vec)))
    eff_col_dist = col_dist / cell_size
    if(dist > 0.708 + eff_col_dist):
        return numpy.array([0.0, 0.0], dtype="float32")
    if(abs(dist_vec[0]) < 0.5 and abs(dist_vec[1]) < 0.5):
        return numpy.float32(0.50 / max(dist, 1.0e-6) * (0.708 + eff_col_dist - dist) * cell_size) * dist_vec
    x_pos = (dist_vec[0] + dist_vec[1] > 0)
    y_pos = (dist_vec[1] - dist_vec[0] > 0)
    if(x_pos and y_pos):
        dist, np = nearest_point(dist_vec, OFFSET_10, OFFSET_01)
    elif((not x_pos) and y_pos):
        dist, np = nearest_point(dist_vec, OFFSET_01, OFFSET_m0)
    elif((not x_pos) and (not y_pos)):
        dist, np = nearest_point(dist_vec, OFFSET_m0, OFFSET_0m)
    elif(x_pos and (not y_pos)):
        dist, np = nearest_point(dist_vec, OFFSET_0m, OFFSET_10)

    if(eff_col_dist < dist):
        return numpy.array([0.0, 0.0], dtype="float32")
    else:
        ori = dist_vec - np
        ori_norm = numpy.sqrt(numpy.sum(ori * ori))
        ori *= 1.0 / max(1.0e-6, ori_norm)
        return numpy.float32(0.50 * (eff_col_dist - dist) * cell_size) * ori

@njit(cache=True)
def vector_move(ori, turn_rate, walk_speed, dt):
    fin_ori = ori + turn_rate * dt
    offset_ori = 0.5 * (fin_ori + ori)
    offset = walk_speed * dt
    d_x = numpy.cos(offset_ori) * offset
    d_y = numpy.sin(offset_ori) * offset
    while(fin_ori > t_PI):
        fin_ori -= t_PI
    while(fin_ori < 0):
        fin_ori += t_PI
    return fin_ori, numpy.asarray([d_x, d_y], dtype="float32")

def vector_move_with_collision(ori, pos, turn_rate, walk_speed, deta_t, cell_walls, cell_size, col_dist):
    slide_factor = 0.20
    if(walk_speed < 0):
        walk_speed *= 0.50
    tmp_pos = numpy.copy(numpy.array(pos, dtype="float32"))
    for _ in range(int(100 * deta_t)):
        ori, offset = vector_move(ori, turn_rate, walk_speed, 0.01)
        exp_pos = tmp_pos + offset
        exp_cell = exp_pos / cell_size
        
        #consider the collision in new cell
        col_f = numpy.array([0.0, 0.0], dtype="float32")
        for i in range(-1, 2): 
            for j in range(-1, 2): 
                w_i = i + int(exp_cell[0])
                w_j = j + int(exp_cell[1])
                if(w_i > -1 and w_i < cell_walls.shape[0] and w_j > -1  and w_j < cell_walls.shape[1]):
                    if(cell_walls[w_i,w_j] > 0):
                        cell_deta = exp_cell - numpy.floor(exp_cell) - numpy.array([i + 0.5, j + 0.5], dtype="float32")
                        col_f += collision_force(cell_deta, cell_size, col_dist)
        tmp_pos = col_f + exp_pos
    return ori, tmp_pos



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
import time
from pygame import font
from numpy import random as npyrnd
from numpy.linalg import norm
from l3c.mazeworld.envs.ray_caster_utils import landmarks_rgb, landmarks_color
from l3c.mazeworld.envs.dynamics import PI

class MazeBase(object):
    def __init__(self, **kw_args):
        for k in kw_args:
            self.__dict__[k] = kw_args[k]
        pygame.init()

    def set_task(self, task_config):
        # initialize textures
        self._cell_walls = numpy.copy(task_config.cell_walls)
        self._cell_texts = task_config.cell_texts
        self._start = task_config.start
        self._n = numpy.shape(self._cell_walls)[0]
        self._cell_landmarks = task_config.cell_landmarks
        self._cell_size = task_config.cell_size
        self._wall_height = task_config.wall_height
        self._agent_height = task_config.agent_height
        self._step_reward = task_config.step_reward
        self._goal_reward = task_config.goal_reward
        self._landmarks_rewards = task_config.landmarks_rewards
        self._landmarks_coordinates = task_config.landmarks_coordinates
        self._landmarks_refresh_interval = task_config.landmarks_refresh_interval
        self._commands_sequence = task_config.commands_sequence
        self._max_life = task_config.max_life
        self._initial_life = task_config.initial_life
        self._int_max = 100000000

        assert self._agent_height < self._wall_height and self._agent_height > 0, "the agent height must be > 0 and < wall height"
        assert self._cell_walls.shape == self._cell_texts.shape, "the dimension of walls must be equal to textures"
        assert self._cell_walls.shape[0] == self._cell_walls.shape[1], "only support square shape"

    def refresh_command(self):
        """
        Update the command for selecting the target to navigate
        At the same time, update the instant_rewards
        Valid only for ``NAVIGATION`` mode
        """
        if(self.task_type != "NAVIGATION"):
            return
        if(self._command is not None):
            x,y = self._landmarks_coordinates[self._command]
            self._instant_rewards[x, y] = 0.0

        self._commands_sequence_idx += 1
        if(self._commands_sequence_idx > len(self._commands_sequence) - 1):
            return True
        self._command = self._commands_sequence[self._commands_sequence_idx]
        x,y = self._landmarks_coordinates[self._command]
        self._instant_rewards[x,y] = self._goal_reward
        return False

    def reach_goal(self):
        g_x, g_y = self._landmarks_coordinates[self._command]
        goal = ((g_x == self._agent_grid[0]) and (g_y == self._agent_grid[1]))
        return goal

    def refresh_landmark_attr(self):
        """
        Refresh the landmarks
            refresh the instant rewards in SURVIVAL mode
            refresh the view in SURVIVAL mode
            No need to refresh for NAVIGATION mode
        """
        if(self.task_type != "SURVIVAL"):
            return
        self._instant_rewards = numpy.zeros_like(self._cell_landmarks, dtype="float32")
        self._cell_active_landmarks = numpy.copy(self._cell_landmarks)
        idxes = numpy.argwhere(self._landmarks_refresh_countdown <= self._landmarks_refresh_interval)
        for idx, in idxes:
            x,y = self._landmarks_coordinates[idx]
            self._cell_active_landmarks[x,y] = -1
        for i, (x,y) in enumerate(self._landmarks_coordinates):
            if(self._cell_active_landmarks[x,y] > -1):
                self._instant_rewards[(x,y)] = self._landmarks_rewards[i]

    def reset(self):
        self._agent_grid = numpy.copy(self._start)
        self._agent_loc = self.get_cell_center(self._start)
        self._agent_trajectory = [numpy.copy(self._agent_grid)]

        # Record all observed cells
        self._cell_exposed = numpy.zeros_like(self._cell_walls).astype(numpy.bool)

        # Maximum w and h in the space
        self._size = self._n * self._cell_size

        # Valid in 3D
        self._agent_ori = 0.0
        self._instant_rewards = numpy.zeros_like(self._cell_landmarks, dtype="float32")
        self._landmarks_refresh_countdown = numpy.full(self._landmarks_rewards.shape, self._int_max)

        # Initialization related to tasks
        if(self.task_type == "SURVIVAL"):
            self._life = self._initial_life
            for i, (x,y) in enumerate(self._landmarks_coordinates):
                self._instant_rewards[(x,y)] = self._landmarks_rewards[i]
            self.refresh_landmark_attr()
        elif(self.task_type == "NAVIGATION"):
            self._commands_sequence_idx = -1
            self._command = None
            self.refresh_command()
        else:
            raise Exception("No such task type: %s" % self.task_type)

        self.update_observation()
        self.steps = 0
        return self.get_observation()

    def evaluation_rule(self):
        self.steps += 1
        self._agent_trajectory.append(numpy.copy(self._agent_grid))
        agent_grid_idx = tuple(self._agent_grid)

        # Landmarks refresh countdown update
        self._landmarks_refresh_countdown -= 1
        idxes = numpy.argwhere(self._landmarks_refresh_countdown <= 0)
        for idx, in idxes:
            self._landmarks_refresh_countdown[idx] = self._int_max

        # Refresh landmarks in SURVIVAL mode, including call back those have been resumed
        if(self.task_type == "SURVIVAL"):
            self.refresh_landmark_attr()

        if(self.task_type == "SURVIVAL"):
            reward = self._instant_rewards[agent_grid_idx] # In survival mode, the step reward is not counted
            self._life += self._step_reward
            self._life = min(self._instant_rewards[agent_grid_idx] + self._life, self._max_life)
            landmark_id = self._cell_landmarks[agent_grid_idx]
            if(landmark_id >= 0 and self._landmarks_refresh_countdown[landmark_id] > self._landmarks_refresh_interval):
                 self._landmarks_refresh_countdown[landmark_id] = self._landmarks_refresh_interval
            done = self._life < 0.0 or self.episode_steps_limit()
        elif(self.task_type == "NAVIGATION"):
            reward = self._instant_rewards[agent_grid_idx] + self._step_reward
            done = False
            if(self.reach_goal()):
                done = self.refresh_command()
            done = done or self.episode_steps_limit()

        return reward, done

    def do_action(self, action):
        raise NotImplementedError()

    def render_init(self, view_size):
        """
        Initialize a God View With Landmarks
        """
        font.init()
        self._font = font.SysFont("Arial", 18)

        #Initialize the agent drawing
        self._render_cell_size = view_size / self._n
        self._view_size = view_size

        self._obs_logo = self._font.render("Observation", 0, pygame.Color("red"))

        self._screen = pygame.Surface((3 * view_size, view_size))
        self._screen = pygame.display.set_mode((3 * view_size, view_size))
        pygame.display.set_caption("MazeWorldRender")
        self._surf_god = pygame.Surface((view_size, view_size))
        self._surf_god.fill(pygame.Color("white"))
        self._surf_lm = pygame.Surface((view_size, view_size))
        self._surf_lm.fill(pygame.Color("grey"))
        it = numpy.nditer(self._cell_walls, flags=["multi_index"])
        for _ in it:
            x,y = it.multi_index
            landmarks_id = self._cell_landmarks[x,y]
            if(self._cell_walls[x,y] > 0):
                pygame.draw.rect(self._surf_god, pygame.Color("black"), (x * self._render_cell_size, y * self._render_cell_size,
                        self._render_cell_size, self._render_cell_size), width=0)
        logo_god = self._font.render("God View - Invisible To Agent", 0, pygame.Color("red"))
        self._surf_god.blit(logo_god,(90, 5))
        logo_loc = self._font.render("Local Map - Invisible To Agent", 0, pygame.Color("red"))
        self._surf_lm.blit(logo_loc,(90, 5))

    def render_godview_dyna(self, scr, offset):
        """
        Cover landmarks with white in case it is not refreshed
        """
        for landmarks_id, (x,y) in enumerate(self._landmarks_coordinates):
            if(self._landmarks_refresh_countdown[landmarks_id] > self._landmarks_refresh_interval):
                pygame.draw.rect(scr, landmarks_color(landmarks_id), 
                        (x * self._render_cell_size + offset[0], y * self._render_cell_size + offset[1],
                        self._render_cell_size, self._render_cell_size), width=0)
        if(self.task_type == "SURVIVAL"):
            txt_life = self._font.render("Life: %f"%self._life, 0, pygame.Color("green"))
            scr.blit(txt_life,(offset[0] + 90, offset[1] + 10))

    def render_localmap(self, scr, offset):
        """
        Cover landmarks with white in case it is not refreshed
        """
        empty_range = 90
        lm_surf = pygame.surfarray.make_surface(self.get_loc_map(3))
        lm_surf = pygame.transform.scale(lm_surf, (self._view_size - 2 * empty_range, self._view_size - 2 * empty_range))
        scr.blit(lm_surf, (offset[0] + empty_range, offset[1] + empty_range))

    def render_observation(self):
        """
        Need to implement the logic for observation painting
        """
        raise NotImplementedError()

    def render_update(self):
        #Paint God View
        self._screen.blit(self._surf_god, (self._view_size, 0))
        self._screen.blit(self._surf_lm, (2 * self._view_size, 0))
        self.render_godview_dyna(self._screen, (self._view_size, 0))
        self.render_localmap(self._screen, (2 * self._view_size, 0))

        #Paint Agent and Observation
        self.render_observation()

        pygame.display.update()
        done = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done=True
        keys = pygame.key.get_pressed()
        return done, keys

    def render_trajectory(self, file_name, additional=None):
        # Render god view with record on the trajectory
        if(additional is not None):
            aw, ah = (additional["surfaces"][0].get_width(),additional["surfaces"][0].get_height())
        else:
            aw, ah = (0, 0)

        traj_screen = pygame.Surface((self._view_size + aw, max(self._view_size, ah)))
        traj_screen.fill(pygame.Color("white"))
        traj_screen.blit(self._surf_god, (0, 0))

        if(self.task_type == "SURVIVAL"):
            self.render_godview_dyna(traj_screen, (0, 0))

        for i in range(len(self._agent_trajectory)-1):
            factor = i / len(self._agent_trajectory)
            noise = (factor - 0.5) * 0.10
            p = self._agent_trajectory[i]
            n = self._agent_trajectory[i+1]
            p = [(p[0] + 0.5 + noise) * self._render_cell_size, (p[1] + 0.5 + noise) *  self._render_cell_size]
            n = [(n[0] + 0.5 + noise) * self._render_cell_size, (n[1] + 0.5 + noise) *  self._render_cell_size]
            pygame.draw.line(traj_screen, pygame.Color(int(255 * factor), int(255 * (1 - factor)), 0, 255), p, n, width=2)

        for landmarks_id, (x,y) in enumerate(self._landmarks_coordinates):
            pygame.draw.rect(traj_screen, landmarks_color(landmarks_id), 
                    (x * self._render_cell_size, y * self._render_cell_size,
                    self._render_cell_size, self._render_cell_size), width=0)

        # paint some additional surfaces where necessary
        if(additional != None):
            for i in range(len(additional["surfaces"])):
                traj_screen.blit(additional["surfaces"][i], (self._view_size, 0))
                pygame.image.save(traj_screen, file_name.split(".")[0] + additional["file_names"][i] + ".png")
        else:
            pygame.image.save(traj_screen, file_name)

    def episode_steps_limit(self):
        return self.steps > self.max_steps-1

    def get_cell_center(self, cell):
        p_x = cell[0] * self._cell_size + 0.5 * self._cell_size
        p_y = cell[1] * self._cell_size + 0.5 * self._cell_size
        return [p_x, p_y]

    def get_loc_grid(self, loc):
        p_x = int(loc[0] / self._cell_size)
        p_y = int(loc[1] / self._cell_size)
        return [p_x, p_y]

    def get_loc_grid_float(self, loc):
        p_x = (loc[0] / self._cell_size)
        p_y = (loc[1] / self._cell_size)
        return [p_x, p_y]

    def movement_control(self, keys):
        """
        Implement the movement control logic, or ''agent dynamics''
        """
        raise NotImplementedError()

    def update_observation(self):
        """
        Update the observation, which is used for returning the state when ''get_observation''
        """
        raise NotImplementedError()

    def get_observation(self):
        return numpy.copy(self._observation)

    def get_loc_map(self, map_range):
        #Add the ground first
        #Find Relative Cells
        x_s = self._agent_grid[0] - map_range
        x_e = self._agent_grid[0] + map_range + 1
        y_s = self._agent_grid[1] - map_range
        y_e = self._agent_grid[1] + map_range + 1
        size = 2 * map_range + 1
        i_s = 0
        i_e = size
        j_s = 0
        j_e = size
        if(x_s < 0):
            i_s = -x_s
            x_s = 0
        if(x_e > self._n):
            i_e -= x_e - self._n
            x_e = self._n
        if(y_s < 0):
            j_s = -y_s
            y_s = 0
        if(y_e > self._n):
            j_e -= y_e - self._n
            y_e = self._n

        # local map: 
        #    ## =-1 for walls
        #    ## >0 for landmarks
        #    ## =0 for empty grounds
        loc_map = - numpy.ones(shape=(2 * map_range + 1, 2 * map_range + 1), dtype="float32")
        loc_map[i_s:i_e, j_s:j_e] = -self._cell_walls[x_s:x_e, y_s:y_e]
        if(self.task_type == "SURVIVAL"):
            loc_map[i_s:i_e, j_s:j_e] += self._cell_active_landmarks[x_s:x_e, y_s:y_e] + 1 # +1 for cell_active_landmarks in [-1, 0~n]
        else:
            loc_map[i_s:i_e, j_s:j_e] += self._cell_landmarks[x_s:x_e, y_s:y_e] + 1 # +1 for cell_active_landmarks in [-1, 0~n]

        loc_map = numpy.expand_dims(loc_map, axis=-1)
        wall_rgb = numpy.array([0, 0, 0], dtype="int32")
        empty_rgb = numpy.array([255, 255, 255], dtype="int32")

        color_map = ((loc_map == -1).astype("int32") * wall_rgb + 
                (loc_map == 0).astype("int32") * empty_rgb)
        for i in landmarks_rgb:
            color_map += (loc_map == (i + 1)).astype("int32") * landmarks_rgb[i].astype("int32")

        if("_agent_ori" in self.__dict__):
            ori = int(2.0 * self._agent_ori / PI + 0.5) % 4 # 0, 1, 2, 3
        else:
            ori = 0

        return numpy.rot90(color_map, k=ori, axes=(1,0)) # Need to rotate color map according to the orientation



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
from collections import namedtuple
from numpy import random as npyrnd
from numpy.linalg import norm

class MazeTaskManager(object):

    # Configurations that decides a specific task
    TaskConfig = namedtuple("TaskConfig", ["start", "cell_landmarks", "cell_walls", "cell_texts", 
        "cell_size", "wall_height", "agent_height", "initial_life", "max_life",
        "step_reward", "goal_reward", "landmarks_rewards", "landmarks_coordinates", "landmarks_refresh_interval", "commands_sequence"])

    def __init__(self, texture_dir, verbose=False):
        pathes = os.path.split(os.path.abspath(__file__))
        texture_dir = os.sep.join([pathes[0], texture_dir])
        texture_files = os.listdir(texture_dir)
        texture_files.sort()
        grounds = [None]
        for file_name in texture_files:
            if(file_name.find("wall") >= 0):
                grounds.append(pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name]))))
            if(file_name.find("ground") >= 0):
                grounds[0] = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
            if(file_name.find("ceil") >= 0):
                self.ceil = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
            if(file_name.find("arrow") >= 0):
                self.arrow = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
        self.grounds = numpy.asarray(grounds, dtype="float32")
        self.verbose = verbose

    @property
    def n_texts(self):
        return self.grounds.shape[0]

    def sample_task(self,
            n=15, 
            allow_loops=True, 
            cell_size=2.0, 
            wall_height=3.2, 
            agent_height=1.6,
            step_reward=-0.01,
            goal_reward=None,
            initial_life=1.0,
            max_life=3.0,
            landmarks_refresh_interval=200,
            landmarks_avg_reward=0.60,
            landmarks_number=5,
            commands_sequence=200,
            wall_density=0.40,
            verbose=False):
        # Initialize the maze ...
        assert n > 6, "Minimum required cells are 7"
        assert n % 2 != 0, "Cell Numbers can only be odd"
        assert landmarks_number > 1, "There must be at least 1 goal, thus landmarks_number must > 1"
        if(landmarks_number > 15):
            landmarks_number = 15
            print("landmarks number too much, set to 15")
        if(self.verbose):
            print("Generating an random maze of size %dx%d, with allow loops=%s, crowd ratio=%f"%(n, n, allow_loops, wall_density))
        cell_walls = numpy.ones(shape=(n, n), dtype="int32")
        cell_texts = numpy.random.randint(1, self.n_texts, size=(n, n))

        # Dig the initial holes
        for i in range(1, n, 2):
            for j in range(1, n, 2):
                cell_walls[i,j] = 0

        #Initialize the logics for prim based maze generation
        wall_dict = dict()
        path_dict = dict()
        rev_path_dict = dict()
        path_idx = 0
        for i in range(1, n - 1):
            for j in range(1, n - 1):
                if(cell_walls[i,j] > 0): # we will keep the axial point
                    wall_dict[i, j] = 0
                elif(cell_walls[i,j] == 0):
                    path_dict[i, j] = path_idx
                    rev_path_dict[path_idx] = [(i,j)]
                    path_idx += 1

        #Prim the wall until all points are connected
        max_cell_walls = numpy.product(cell_walls[1:-1, 1:-1].shape)
        while len(rev_path_dict) > 1 or (allow_loops and numpy.sum(cell_walls[1:-1, 1:-1]) > max_cell_walls * wall_density):
            wall_list = list(wall_dict.keys())
            random.shuffle(wall_list)
            for i, j in wall_list:
                new_path_id = -1
                connected_path_id = dict()
                abandon_path_id = dict()
                max_duplicate = 1

                for d_i, d_j in [(i - 1, j), (i + 1, j), (i, j - 1), (i, j + 1)]:
                    if((d_i > 0  and d_i < n and d_j > 0 and d_j < n)
                            and cell_walls[d_i, d_j] < 1):
                        # calculate duplicate path id that might creat a loop
                        if path_dict[d_i, d_j] not in connected_path_id:
                            connected_path_id[path_dict[d_i, d_j]] = 1
                        else:
                            connected_path_id[path_dict[d_i, d_j]] += 1
                        if(connected_path_id[path_dict[d_i, d_j]] > max_duplicate):
                            max_duplicate = connected_path_id[path_dict[d_i, d_j]]

                        # decide the new path_id and find those to be deleted
                        if(path_dict[d_i, d_j] < new_path_id or new_path_id < 0):
                            if(new_path_id >= 0):
                                abandon_path_id[new_path_id] = (new_i, new_j)
                            new_path_id = path_dict[d_i, d_j]
                            new_i = d_i
                            new_j = d_j
                        elif(path_dict[d_i, d_j] != new_path_id): # need to be abandoned
                            abandon_path_id[path_dict[d_i, d_j]] = (d_i, d_j)
                if(len(abandon_path_id) >= 1 and max_duplicate < 2):
                    break
                if(len(abandon_path_id) >= 1 and max_duplicate > 1 and allow_loops):
                    break
                if(allow_loops and len(rev_path_dict) < 2 and random.random() < 0.2):
                    break

            if(new_path_id < 0):
                continue
                        
            # add the released wall
            rev_path_dict[new_path_id].append((i,j))
            path_dict[i,j] = new_path_id
            cell_walls[i,j] = 0
            del wall_dict[i,j]

            # merge the path
            for path_id in abandon_path_id:
                rev_path_dict[new_path_id].extend(rev_path_dict[path_id])
                for t_i_o, t_j_o in rev_path_dict[path_id]:
                    path_dict[t_i_o,t_j_o] = new_path_id
                del rev_path_dict[path_id]

        #Paint the texture of passways to ground textures 
        for i in range(1, n - 1):
            for j in range(1, n - 1):
                if(cell_walls[i,j] < 1):
                    cell_texts[i,j] = 0

        #Randomize a start point and n landmarks
        landmarks_likelihood = numpy.random.rand(n, n) - cell_walls
        idxes = numpy.argsort(landmarks_likelihood, axis=None)
        topk_idxes = idxes[-landmarks_number-1:]

        def idx_trans(idx):
            return (idx // n, idx % n)

        start = idx_trans(topk_idxes[-1])
        landmarks = [idx_trans(i) for i in topk_idxes[:-1]]

        #Calculate goal reward, default is - n sqrt(n) * step_reward
        assert step_reward < 0, "step_reward must be < 0"
        if(goal_reward is None):
            def_goal_reward = - numpy.sqrt(n) * n * step_reward
        else:
            def_goal_reward = goal_reward
        assert def_goal_reward > 0, "goal reward must be > 0"

        commands_sequence = numpy.random.randint(0, len(landmarks), commands_sequence)

        # Generate landmark rewards
        landmarks_rewards = numpy.random.rand(landmarks_number) - 0.50 + landmarks_avg_reward
        cell_landmarks = numpy.zeros_like(cell_walls) - 1
        for i,idx in enumerate(landmarks):
            cell_landmarks[tuple(idx)] = int(i)
        cell_landmarks = cell_landmarks.astype(cell_walls.dtype)
        if(verbose):
            print("\n\n---------Successfully generate maze task with the following attributes-----------\n")
            print("Maze size %s x %s" %(n, n)) 
            print("Initialze born location: %s,%s" % start)
            integrate_maze = cell_landmarks + 1 - cell_walls
            print("Maze configuration (-1: walls, 0 empty, >1 landmarks and ID): \n%s" % integrate_maze)
            print("Commands sequence: \n%s" % commands_sequence)
            print("Landmarks Rewards: \n%s" % landmarks_rewards)
            print("Landmarks Refresh Interval: \n%s" % landmarks_refresh_interval)
            print("\n----------------------\n\n")

        return MazeTaskManager.TaskConfig(
                start=start,
                cell_landmarks=cell_landmarks,
                cell_walls=cell_walls,
                cell_texts=cell_texts,
                cell_size=cell_size,
                step_reward=step_reward,
                goal_reward=def_goal_reward,
                wall_height=wall_height,
                agent_height=agent_height,
                initial_life=initial_life,
                max_life=max_life,
                commands_sequence=commands_sequence,
                landmarks_coordinates=landmarks,
                landmarks_rewards=landmarks_rewards,
                landmarks_refresh_interval=landmarks_refresh_interval
                )

MAZE_TASK_MANAGER=MazeTaskManager("img")
MazeTaskSampler = MAZE_TASK_MANAGER.sample_task

if __name__=="__main__":
    task = MazeTaskSampler(verbose=True)



#   Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import _io
import numpy
import gym
from numpy import random

class MetaLangV1(gym.Env):
    """
    Meta Language Model
    MetaLM(V, n, l, e, L) generates n Exponential(1/l)-length sequences, which appears repeatedly until reaching length L
    Each time it repeats, we add a noise by replacing the number with e probablity 
    V: Vocabulary Size
    l: mean repeating length
    e: noise ratio
    L: Sequence Length

    The task require the agent to recover the correct sequence.
    """

    def __init__(self, 
            V=64, 
            n=10, 
            l=64, 
            e=0.10, 
            L=2048):
        self.L = int(L)
        self.V = int(V)
        self.lamb = l
        self.n = n
        self.e = float(e)
        self.mask_ratio = 0.30
        assert n >= 1 and V > 1 and l > 1 and e > 0 and e < 1 and L > 1

    def add_noise(self, seq):
        """
        Add noise to a sequence, return the new sequence
        """
        noise_value = random.randint(0, self.V, size=(numpy.shape(seq)), dtype="int32")
        noise_ratio = (random.random(size=(numpy.shape(seq))) < self.e).astype("int32")
        mask_ratio = (random.random(size=(numpy.shape(seq))) < self.mask_ratio).astype("int32")
        diff = noise_value - seq
        diff = diff * (noise_ratio!=0).astype("int32")
        new_seq = seq + diff
        new_seq = new_seq * (1 -  mask_ratio * noise_ratio)
        return new_seq

    def elements_generator(self, seed=None):
        elements = []
        if(seed is not None):
            numpy.random.seed(seed)
        for _ in range(self.n):
            l_r = max(3, numpy.random.poisson(self.lamb))
            elements.append(random.randint(0, self.V, size=(l_r), dtype="int32"))
        return elements

    def data_generator(self, seed=None):
        elements = self.elements_generator(seed=seed)
        features = []
        labels = []
        cur_l = 0
        while cur_l < self.L + 1:
            seq = elements[random.randint(0, self.n)]
            fea = self.add_noise(seq)
            sep = numpy.array([self.SepID], dtype="int32")
            features.append(fea)
            labels.append(seq)
            features.append(sep)
            labels.append(sep)
            cur_l += len(seq) + 1
        features = numpy.concatenate(features, axis=0).astype("int32")
        labels = numpy.concatenate(labels, axis=0).astype("int32")
        return features[:self.L], labels[1:(self.L+1)]

    def batch_generator(self, batch_size, seed=None):
        features = []
        labels = []
        for _ in range(batch_size):
            fea, lab = self.data_generator(seed=seed)
            features.append(fea.tolist())
            labels.append(lab.tolist())
        features = numpy.asarray(features)
        labels = numpy.asarray(labels)
        return features, labels

    def generate_text(self, size, output_stream):
        features, labels = self.batch_generator(size)
        if(isinstance(output_stream, _io.TextIOWrapper)):
            need_close = False
        elif(isinstance(output_stream, str)):
            output_stream = open(output_stream, "w")
            need_close = True
        for i in range(features.shape[0]):
            output_stream.write("\t".join(map(lambda x: "%d,%d"%(x[0],x[1]), zip(features[i].tolist(), labels[i].tolist()))))
            output_stream.write("\n")
        if(need_close):
            output_stream.close()

    def generate_npy(self, size, file_name):
        feas, labs = self.batch_generator(size)
        numpy.save(file, numpy.stack([feas, labs], axis=0))

    @property
    def VocabSize(self):
        return self.V

    @property
    def SepID(self):
        return 0

    @property
    def MaskID(self):
        return 0

    @property
    def PaddingID(self):
        return 0
